import numpy as np
import math


class Math:
    """
    Class to simplify and encapsulate functionality related
    to math that could be more complex than the contained 
    in the basic python math module.
    """
    @staticmethod
    def sigmoid(value):
        """
        TODO: Write doc about this
        """
        # TODO: Maybe avoid np by now and use math
        return 1.0 / (1 + np.exp(-value))
    
    @staticmethod
    def normalize(value, min_value, max_value):
        """
        Normalize the provided 'value' to turn it into another
        value between 0.0 and 1.0.
        """
        return (value - min_value) / (max_value - min_value)

    @staticmethod
    def denormalize(value, min_value, max_value):
        """
        Denormalize the provided 'value' (that must be between
        0.0 and 1.0) by turning it into another value between
        the provided 'min_value' and 'max_value'.
        """
        return value * (max_value - min_value) + min_value

    @staticmethod
    def logarithm(value, base):
        """
        Calculate the logarithm in 'base' base of the provided
        'value' number.
        """
        return math.log(value, base)

    @staticmethod
    def is_power_of_n(value, base):
        """
        Calculate if the provided 'value' number is a power
        of the also provided 'base' number. This means that
        16 and 4 will return True, and 15 and 4, False.
        """
        return Math.logarithm(value, base).is_integer()

    


from yta_general_utils.file.filename import sanitize_filename, get_filename
from yta_general_utils.path import get_abspath_parent_folder

import inspect
import sys
import os


def get_code_abspath(code):
    """
    Returns the abspath of the file in which the code is written.
    The 'code' parameter must be a module, class, method, function,
    traceback, frame or code object to be correctly inspected.
    """
    return sanitize_filename(inspect.getfile(code))

def get_code_filename(code):
    """
    Returns the filename in which the code is written. The 'code' 
    parameter must be a module, class, method, function, traceback, 
    frame or code object to be correctly inspected.

    This method will include the filename with the extension.
    """
    return get_filename(inspect.getfile(code))

def get_project_abspath():
    """
    Returns the absolute path of the current project (the
    one that is being executed and using this library.

    The absolute path returned ends in '/' and has been
    sanitized.
    """
    return sanitize_filename(os.getcwd()) + '/'

def get_current_file_abspath(parent_levels: int = 0):
    """
    Returns the absolute path of the file that is currently
    being executed (in which the code is written). If 
    'parent_levels' provided, it will return the abspath
    to the parent folder that corresponds to the level
    requested.

    The absolute path is returned ending in '/' and has
    been sanitized.
    """
    abspath = sanitize_filename(os.path.dirname(os.path.abspath(sys.argv[0])))
    abspath = get_abspath_parent_folder(abspath, parent_levels) if parent_levels > 0 else f'{abspath}/'

    return abspath


from yta_general_utils.color.utils import is_hexadecimal_color, parse_rgba_color, rgba_to_hex, hex_to_rgba, rgb_to_hex, rgba_to_hex, rgb_to_hsl, rgb_to_cymk, rgb_to_hsv, hsv_to_rgb, rgba_to_hsv, parse_color, hsv_to_rgba
from yta_general_utils.programming.parameter_validator import PythonValidator
from typing import Union


class ColorConverter:
    """
    Class to simplify and encapsulate the functionality
    related to color conversion.
    """

    @staticmethod
    def rgb_to_hex(red, green, blue):
        """
        Returns the provided RGB color as a hex color. The 'red', 'green' and
        'blue' parameters must be between 0 and 255.
        """
        return rgba_to_hex(red, green, blue)
    
    @staticmethod
    def hex_to_rgb(color: str):
        """
        Parse the provided hexadecimal 'color' parameter and
        turn it into an RGB color (returned as r,g,b) or
        raises an Exception if not.
        """
        r, g, b, _ = ColorConverter.hex_to_rgba(color)

        return r, g, b

    @staticmethod
    def hex_to_rgba(color: str):
        if not is_hexadecimal_color(color):
            raise Exception(f'The provided "color" parameter "{str(color)}" is not an hexadecimal color.')
        
        return hex_to_rgba(color)

    @staticmethod
    def rgb_to_hex(color: Union[tuple, list], do_include_alpha: bool = False):
        """
        Parse the provided RGB 'color' parameter and turn it to
        a hexadecimal color if valid or raises an Exception if
        not. The result will be #RRGGBB if 'do_include_alpha' is
        False, or #RRGGBBAA if 'do_include_alpha' is True.
        """
        validate_color(color)

        return rgb_to_hex(color, do_include_alpha)
        
    @staticmethod
    def rgba_to_hex(color: Union[tuple, list], do_include_alpha: bool = False):
        """
        Parse the provided RGBA 'color' parameter and turn it to
        a hexadecimal color if valid or raises an Exception if
        not. The result will be #RRGGBB if 'do_include_alpha' is
        False, or #RRGGBBAA if 'do_include_alpha' is True.
        """
        validate_color(color)

        return rgba_to_hex(color, do_include_alpha)

    @staticmethod
    def rgba_to_hsl(color: Union[tuple, list]):
        # TODO: Explain
        validate_color(color)

        _, _, _, a = parse_rgba_color(color)
        
        return *ColorConverter.rgb_to_hsl(color), a

    @staticmethod
    def rgb_to_hsl(color: Union[tuple, list]):
        # TODO: Explain
        validate_color(color)
        
        return rgb_to_hsl(color)

    @staticmethod
    def rgba_to_cymk(color: Union[tuple, list]):
        # TODO: Explain
        validate_color(color)

        color = color[:3]

        # TODO: Is there a way to handle alpha transparency
        # with a cymk (?)
        return ColorConverter.rgb_to_cymk(color)

    @staticmethod
    def rgb_to_cymk(color: Union[tuple, list]):
        # TODO: Explain
        # It looks like you need to know the color profile before
        # any conversion from RGB or RGBA
        # https://www.reddit.com/r/AdobeIllustrator/comments/17vpbod/different_cmyk_values_for_same_hex_code_which/?rdt=55781#:~:text=A%20hex%20code%20is%20an,information%2C%20like%20a%20colour%20profile.
        validate_color(color)

        return rgb_to_cymk(color)
    
    @staticmethod
    def rgb_to_hsv(color: Union[tuple, list]):
        """
        Turn the provided RGB 'color' into a HSV color.
        """
        validate_color(color)

        return rgb_to_hsv(color)

    @staticmethod
    def rgba_to_hsv(color: Union[tuple, list]):
        """
        Turn the provided RGBA 'color' into a HSV color.
        The HSV color doesn't pay attention to the alpha
        layer so this method is the same as using the
        'rgb_to_hsv' method for this 'color'.
        """
        validate_color(color)

        return rgba_to_hsv(color)
    
    @staticmethod
    def hsv_to_rgb(color: Union[tuple, list]):
        """
        Turn the provided HSV 'color' into a RGB color.
        """
        validate_color(color)
        
        h, s, v = color

        return hsv_to_rgb(h, s, v)
    
    @staticmethod
    def hsv_to_rgba(color: Union[tuple, list]):
        """
        Turn the provided HSV 'color' into a RGBA color.
        The HSV color doesn't pay attention to the alpha
        layer so it will be always one.
        """
        validate_color(color)

        h, s, v = color

        return hsv_to_rgba(h, s, v)
    
def validate_color(color: Union[tuple, list]):
    """
    Validate the provided 'color' as a tuple or list of 3
    or 4 elements.
    """
    if not PythonValidator.is_instance(color, 'Color') and parse_color(color) is None:
        raise Exception('The provided "color" is not a parsable color.')
    
    return True


"""
    TODO: This could be mixed with the 'filename' functionality
    as it is similar but we treat it as paths and filenames while
    they are the same (a filename is a relative path)
"""
from yta_general_utils.file.filename import sanitize_filename
from pathlib import Path


def get_abspath_parent_folder(
    abspath: str,
    levels: int = 1
) -> str:
    """
    Iterates over the provided 'abspath' and goes to the parent
    folder that is 'levels' levels above. This method will
    return the new abspath sanitized and ended in '/'.
    """
    # TODO: Handle when 'levels' is greater than possible
    import os.path

    if levels <= 0:
        return abspath

    for _ in range(levels):
        abspath = os.path.dirname(abspath)

    return f'{sanitize_filename(abspath)}/'

def create_file_abspath(
    file_abspath: str
) -> str:
    """
    This method will create the folders needed to be able to
    create the provided 'file_abspath'. This will create
    all the folders until reaching the file level.

    @param
        **file_abspath**
        File absolute path that we need to be able to work with
        that file. This must include filename and extension.
    """
    if not file_abspath or not ':/' in file_abspath or file_abspath.endswith('/'):
        raise Exception('No valid absolute path provided.')

    Path(get_abspath_parent_folder(file_abspath)).mkdir(parents = True, exist_ok = True)

    return file_abspath

def is_abspath(
    filename: str
) -> bool:
    """
    Checks if the provided 'filename' is an absolute path or not,
    returning True if yes or False if not.
    """
    # TODO: Please, improve this with a library or something
    # Check this: https://stackoverflow.com/questions/3320406/how-to-check-if-a-path-is-absolute-path-or-relative-path-in-a-cross-platform-way
    return ':\\' in filename or ':/' in filename



from yta_general_utils.file.writer import FileWriter
from yta_general_utils.dataclasses import FileReturn
from typing import Union

import requests


def get_file(
    url: str,
    output_filename: Union[str, None] = None
):
    """
    This method sends a request to the provided 'url'
    if provided and obtains the file content (if 
    possible). It will write the obtained file locally
    as 'output_filename' if provided.

    This method returns the file content data as 
    obtained from the requested (.content field).
    """
    # # This code below is to write it by chunks, if
    # # the other way fails
    # CHUNK_SIZE = 8192
    
    # content = b''
    # with requests.get(url, stream = True) as response:
    #     response.raise_for_status()
    #     with open(output_filename, 'wb') if output_filename else nullcontext() as file:
    #         for chunk in response.iter_content(CHUNK_SIZE):
    #             content += chunk
    #             if file:
    #                 file.write(chunk)

    # return content

    content = requests.get(url).content

    if output_filename:
        FileWriter.write_binary_file(content, output_filename)

    return content

def download_file(
    url: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Receives a downloadable url as 'url' and downloads that
    file in our system as 'output_filename'.
    """
    if output_filename is None:
        raise Exception('No "output_filename" provided.')
    
    file = get_file(url, output_filename)
    
    return FileReturn(
        file,
        output_filename
    )


from yta_general_utils.checker.url import verify_image_url
from yta_general_utils.downloader.utils import download_file
from yta_general_utils.programming.output import Output
from yta_general_utils.dataclasses import FileReturn
from yta_general_utils.file.enums import FileType
from yta_general_utils.programming.parameter_validator import PythonValidator
from PIL import Image
from typing import Union

import requests


def download_image(
    url: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Download the image from the provided 'url' and stores it
    locally as 'output_filename' if provided, or as a
    temporary file if not.
    
    This method sends two requests. The first one is to check
    if the provided 'url' contains a valid image, and the
    second one is to download it.

    TODO: Maybe rename to 'download_with_check'.
    """
    image_extension = verify_image_url(url)
    if not image_extension:
        raise Exception('Url "' + url + '" is not a valid image url.')

    image_extension = image_extension.replace('.', '')

    # TODO: Maybe we want to return the content instead of the filename
    # so, if they don't provide 'output_filename' we could return the
    # content instead, and only download it if 'output_filename' is 
    # provided (fixing it if is wrong) and return the final 
    # 'output_filename' in this last case.
    return download_file(url, Output.get_filename(output_filename, image_extension))

def download_image_2(
    url: str,
    output_filename: str
) -> FileReturn:
    """
    Download the image from the provided 'url' and store it
    locally as 'output_filename'.

    This method doesn't check if the url contains a valid 
    image, so the only request done is the one which downloads
    it.

    TODO: Maybe rename to 'download_without_check'.
    """
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" is not a valid string.')
    
    output_filename = Output.get_filename(output_filename, FileType.IMAGE)

    import shutil
    
    res = requests.get(url, stream = True)

    if res.status_code == 200:
        with open(output_filename, 'wb') as f:
            shutil.copyfileobj(res.raw, f)
    else:
        raise Exception(f'Something went wrong when trying to download the image from "{url}".')

    return FileReturn(
        Image.open(output_filename, mode = 'RGBA'),
        output_filename
    )


from yta_general_utils.programming.parameter_validator import NumberValidator

import random
import string


class Random:
    """
    Class to wrap our custom random methods.
    """

    @staticmethod
    def float_between(
        start: float,
        end: float,
        step: float
    ) -> float:
        """
        Get a random float number in between the range
        by the given 'start' and 'end' limits, using the
        also provided 'step'.

        The limits are inverted if the 'end' is lower
        than the 'start'.

        TODO: Is the limit included (?) Review and, if
        necessary, include it as a parameter.
        """
        if not NumberValidator.is_number(start):
            raise Exception('The provided "start" parameter is not a number.')
        
        if not NumberValidator.is_number(end):
            raise Exception('The provided "end" parameter is not a number.')
        
        if not NumberValidator.is_number(step):
            raise Exception('The provided "step" parameter is not a number.')
        
        start = float(start)
        end = float(end)
        step = float(step)

        # TODO: What about step = 0 or things like that (?)
        # TODO: What if 'start' and 'end' are the same (?)

        # Swap limits if needed
        start, end = end, start if end < start else start, end

        return random.choice([
            round(start + i * step, 4)
            for i in range(int((end - start) / step) + 1)
            if start + i * step <= end]
        )
    
    @staticmethod
    def int_between(
        start: int,
        end: int,
        step: int = 1 
    ) -> int:
        """
        Get a random int number in between the range
        by the given 'start' and 'end' limits, using the
        also provided 'step'.

        The limits are inverted if the 'end' is lower
        than the 'start'. The limist are included.
        """
        # TODO: What about strings that are actually parseable
        # as those numbers (?)
        if not NumberValidator.is_number(start):
            raise Exception('The provided "start" parameter is not a number.')
        
        if not NumberValidator.is_number(end):
            raise Exception('The provided "end" parameter is not a number.')
        
        if not NumberValidator.is_number(step):
            raise Exception('The provided "step" parameter is not a number.')
        
        start = int(start)
        end = int(end)
        step = int(step)

        # TODO: What about step = 0 or things like that (?)
        # TODO: What if 'start' and 'end' are the same (?)
        
        # Swap limits if needed
        start, end = end, start if end < start else start, end
        
        return random.randrange(start, end + 1, step)
    
    @staticmethod
    def bool():
        """
        Get a boolean value chosen randomly.
        """
        return bool(random.getrandbits(1))
    
    @staticmethod
    def characters(n: int = 10):
        """
        Get a string with 'n' random characters.
        """
        return ''.join(random.choices(string.ascii_letters, k = n))
    
    @staticmethod
    def digits(n: int = 10):
        """
        Get a string with 'n' random digits.
        """
        return ''.join(random.choices(string.digits, k = n))
    
    @staticmethod
    def characters_and_digits(n: int = 10):
        """
        Get a string with 'n' random characters and digits.
        """
        ''.join(random.choices(string.ascii_letters + string.digits, k = n))


# TODO: Find a more specific and better way to
# classify these files and methods. I moved here
# to avoid using __init__.py
from yta_general_utils.file.filename import sanitize_filename
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from shutil import move as shutil_move, copyfile as shutil_copyfile

import glob


class FileSearchOption(Enum):
    """
    Enum that allows us setting the strategy dynamically when 
    searching for files.
    """

    FILES_AND_FOLDERS = 'fifo'
    """
    This option, when set, will return files and folders.
    """
    FILES_ONLY = 'fi'
    """
    This option, when set, will return files only.
    """
    FOLDERS_ONLY = 'fo'
    """
    This option, when set, will return folders only.
    """

class FileHandler:
    """
    Class to simplify the way we handle files and folders.
    """

    @staticmethod
    def get_list(
        abspath: str,
        option: FileSearchOption = FileSearchOption.FILES_AND_FOLDERS,
        pattern: str = '*',
        recursive: bool = False
    ):
        """
        List what is inside the provided 'abspath'. This method will list files and
        folders, files or only folders attending to the provided 'option'. It will
        also filter the files/folders that fit the provided 'pattern' (you can use
        '*' as wildcard, so for example '*.jpg' will list all images). This method
        can also be used in a recursive way if 'recursive' parameter is True, but
        take care of memory consumption and it would take its time to perform.

        This method returns a list with all existing elements absolute paths 
        sanitized.
        """
        if not abspath:
            return None
        
        abspath = sanitize_filename(abspath)

        # This below get files and folders
        files_and_folders = [
            sanitize_filename(f)
            for f in glob.glob(pathname = abspath + pattern, recursive = recursive)
        ]

        return {
            FileSearchOption.FILES_ONLY: [
                f
                for f in files_and_folders
                if FileValidator.is_file(f)
            ],
            FileSearchOption.FOLDERS_ONLY: [
                f
                for f in files_and_folders
                if FileValidator.is_folder(f)
            ],
            FileSearchOption.FILES_AND_FOLDERS: files_and_folders
        }[option]

    @staticmethod
    def rename_file(origin_filename: str, destination_filename: str, replace_if_existing: bool = False):
        """
        Renames the 'origin_filename' to the 'destination_filename'.
        If 'replace_if_existing' is True, it will replace the destination
        file if possible and allowed. If it is False, it will fail.

        TODO: Remove 'replace_if_existing' if not used.
        """
        if not origin_filename:
            return None
        
        if not destination_filename:
            return None
        
        shutil_move(origin_filename, destination_filename)

    @staticmethod
    def copy_file(origin_filename: str, destination_filename: str):
        """
        Makes a copy of the provided 'origin_filename' and 
        stores it as 'destination_filename'.

        The destination folder must exist.
        """
        if not origin_filename:
            return None

        if not destination_filename:
            return None

        shutil_copyfile(origin_filename, destination_filename)


from yta_general_utils.programming.parameter_validator import PythonValidator
from urllib import parse as url_parse


# TODO: Refactor this
def encode_url_parameter(parameter: str):
    """
    Encode the provided 'parameter' to be able to send
    it throught as a url query parameter.

    This method will turn 'example message' to
    'example%20message'.
    """
    if not PythonValidator.is_string(parameter):
        raise Exception('The provided "parameter" parameter is not a string.')
    
    return url_parse.quote(parameter)

def encode_url_parameters(parameters: dict):
    """
    Encode the 'parameters' provided as a dict to be
    able to send them through a url query.

    This method will concatenate the parameters as
    '?query=esto%20es%20una%20frase&page=3...'.
    """
    if not PythonValidator.is_dict(parameters):
        raise Exception('The provided "parameters" parameter is not a dict.')
    
    return url_parse.urlencode(parameters)


from yta_general_utils.color.utils import parse_color
from yta_general_utils.color.converter import ColorConverter
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.color.enums import ColorString
from typing import Union


class Color:
    """
    Class that represents a color, stored as RGBA, makes
    easy the way we interact with color and provide them as
    parameters and simplify the color conversion. The color
    is stored as a not-normalized color, but values can be
    normalized through the methods that allow it (those 
    including the 'normalized' bool parameter).

    Any attribute has to be initialized with a value between
    0 and 255. Alpha as 255 means full opaque.

    TODO: Please confirm alpha 255 is opaque.
    """

    r: int
    """
    Red color, from 0 to 255, where 0 is no value and 255
    is everything.
    """
    g: int
    """
    Green color, from 0 to 255, where 0 is no value and 255
    is everything.
    """
    b: int
    """
    Blue color, from 0 to 255, where 0 is no value and 255
    is everything.
    """
    a: int
    """
    Alpha (transparency), from 0 to 255, where 0 is no
    value and 255 is everything.
    """

    def __init__(self, r, g, b, a):
        self.r, self.g, self.b, self.a = r, g, b, a
        # TODO: Precalculate and store all the values 
        # internally so we don't need to calculate them
        # later (?)

    @property
    def rgb_not_normalized(self):
        """
        Get the color as a tuple of the 3 RGB values that
        are, in order: red, green, blue. These values are
        not normalized, so each value is in the interval
        [0, 255].
        """
        return self.r, self.g, self.b
    
    @property
    def rgb_normalized(self):
        """
        Get the color as a tuple of the 3 RGB values that
        are, in order: red, green, blue. These values are
        normalized, so each value is in the interval [0, 1].
        """
        return self.r / 255.0, self.g / 255.0, self.b / 255.0
    
    @property
    def rgba_not_normalized(self):
        """
        Get the color as a tuple of the 3 RGB values and
        a 4th value representing the transparency, that
        are, in order: red, green, blue, alpha. These
        values are not normalized, so each value is in
        the interval [0, 255].
        """
        self.r, self.g, self.b, self.a

    @property
    def rgba_normalized(self):
        """
        Get the color as a tuple of the 3 RGB values and
        a 4th value representing the transparency, that
        are, in order: red, green, blue, alpha. These
        values are normalized, so each value is in the
        interval [0, 1].
        """
        return self.r / 255.0, self.g / 255.0, self.b / 255.0, self.a / 255.0
    
    @property
    def rgb_array_not_normalized(self):
        """
        Get the color as an array of the 3 RGB values that
        are, in order: red, green, blue. These values are
        not normalized, so each value is in the interval
        [0, 255].
        """
        return [*self.rgba_not_normalized]
    
    @property
    def rgb_array_normalized(self):
        """
        Get the color as an array of the 3 RGB values that
        are, in order: red, green, blue. These values are
        normalized, so each value is in the interval [0, 1].
        """
        return [*self.rgba_normalized]
    
    @property
    def rgba_array_not_normalized(self):
        """
        Get the color as an array of the 3 RGB values and
        a 4th value representing the transparency, that
        are, in order: red, green, blue, alpha. These
        values are not normalized, so each value is in
        the interval [0, 255].
        """
        return [*self.rgba_not_normalized]
    
    @property
    def rgba_array_normalized(self):
        """
        Get the color as an array of the 3 RGB values and
        a 4th value representing the transparency, that
        are, in order: red, green, blue, alpha. These
        values are normalized, so each value is in the
        interval [0, 1].
        """
        return [*self.rgba_normalized]
    
    @property
    def hex_with_alpha(self):
        """
        Get the color as a string representing it in
        hexadecimal value. The result will be #RRGGBBAA
        because it includes the alpha value in the last
        position.
        """
        return ColorConverter.rgba_to_hex(self.rgba_not_normalized, True)

    @property
    def hex_without_alpha(self):
        """
        Get the color as a string representing it in
        hexadecimal value. The result will be #RRGGBB
        because it doesn't include the alpha value.
        """
        return ColorConverter.rgba_to_hex(self.rgba_not_normalized, False)
    
    @property
    def hsl(self):
        """
        Get the color as an HSL color.
        """
        return ColorConverter.rgba_to_hsl(self.rgba_not_normalized)
    
    @property
    def cymk(self):
        """
        Get the color as an CYMK color.
        """
        return ColorConverter.rgba_to_cymk(self.rgba_not_normalized)

    @property
    def hsv(self):
        """
        Get the color as a HSV color.
        """
        return ColorConverter.rgba_to_hsv(self.rgba_not_normalized)
    
    # TODO: Use the cv2 library to make other changes
    @staticmethod
    def parse(color: Union[list, tuple, str, 'ColorString', 'Color']):
        """
        Parse the provided 'color' parameter and return the
        color as r,g,b,a values or raises an Exception if it
        is not a valid and parseable color.

        This method accepts string colors (if names are
        registered in our system), hexadecimal colors (than
        can also include alpha value), RGB array or tuples
        (that can be normalized, with float values between
        0.0 and 1.0, or not normalized, with int values
        between 0 and 255), or RGBA array or tuples, similar
        to RGB but including a 4h alpha value.
        """
        if PythonValidator.is_instance(color, Color):
            return color

        color = parse_color(color)

        if color is None:
            raise Exception(f'The provided "color" parameter is not parseable.')
        
        return Color(*color)



from yta_general_utils.web.scrapper.chrome_scrapper import ChromeScrapper
from yta_general_utils.programming.output import Output
from yta_general_utils.dataclasses import FileReturn
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.downloader import Downloader
from selenium.webdriver.common.by import By
from typing import Union


def get_instagram_video(
    url: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Gets the Instagram video (reel) from the provided 'url' (if valid)
    and returns its data or stores it locally as 'output_filename' if
    provided.
    """
    # This method is based on the external website below, so
    # it could stop working when that website is off.
    # TODO: Try to make alternatives with other web pages.
    DOWNLOAD_INSTAGRAM_VIDEO_URL = 'https://downloadgram.org/video-downloader.php'
    
    scrapper = ChromeScrapper()
    scrapper.go_to_web_and_wait_util_loaded(DOWNLOAD_INSTAGRAM_VIDEO_URL)

    # We need to place the url in the input and press enter
    url_input = scrapper.find_element_by_id('url')
    url_input.send_keys(url)

    submit_button = scrapper.find_element_by_id('submit')
    submit_button.click()

    # We need to wait until video is shown
    video_element = scrapper.find_element_by_element_type_waiting('video')
    video_source_element = video_element.find_element(By.TAG_NAME, 'source')
    video_source_url = video_source_element.get_attribute('src')

    # This just downloads the thumbnail but, for what (?)
    # thumbnail_image_url = video_element.get_attribute('poster')
    # download_image(thumbnail_image_url, 'test_instagram_image.png')

    return Downloader.download_video(
        video_source_url,
        Output.get_filename(output_filename, FileTypeX.VIDEO)
    )

# TODO: Implement 'get_instagram_story' (?)

"""
# Check: https://github.com/gabrielkheisa/instagram-downloader/blob/main/run.py
# He downloads with selenium
# This and the one below: https://stackoverflow.com/a/48705202
# This code (https://github.com/instaloader/instaloader/tree/master) is used
# by RocketAPI to charge you
"""


from yta_general_utils.downloader.utils import download_file
from yta_general_utils.dataclasses import FileReturn
from yta_general_utils.file.enums import FileType
from yta_general_utils.programming.output import Output
from typing import Union


def download_audio(
    url: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Receives a downloadable url as 'url' and downloads that audio in
    our system as 'output_filename'.

    This method will return the final 'output_filename' in which the
    file has been downloaded if so.
    """
    return download_file(url, Output.get_filename(output_filename, FileType.AUDIO))


# TODO: Create a ConsolePrinter class maybe (?)
def print_error(message: str = ''):
    print('>>>> [ERROR] <<<< ' + message)

def print_in_progress(message: str = ''):
    print('.... ' + message)

def print_completed(message: str = ''):
    print('[OK] ' + message)


from yta_general_utils.math.value_normalizer import ValueNormalizer
from yta_general_utils.math.rate_functions.rate_function_argument import RateFunctionArgument
from yta_general_utils.programming.parameter_validator import NumberValidator

import numpy as np


class Progression:
    """
    Class to represent a progression of 'n' values from
    'start' to 'end' (including both in that 'n' amount),
    with an associated rate function to calculate the
    values between those 'start' and 'end' limits.

    The progression can be ascending or descending.

    This class is useful to obtain each individual value
    that must be applied in different use cases.
    """
    
    _start: float = None
    _end: float = None
    _n: float = None
    """
    The number of values including 'start' and 'end'. 
    This number must be greater or equal to 2.
    """
    _values: list[float] = None
    """
    The amount of values that will exist the progression
    (including 'start' and 'end'). A progression with
    n = 2 will be as simple as [start, end].
    """
    _rate_function: RateFunctionArgument = None
    """
    The function to calculate the values in between the
    start and end limits.
    """

    @property
    def values(self):
        """
        The list of 'n' values (including 'start' and 'end') 
        between those limits and according to the provided
        'rate_function'.
        """
        if self._values is None:
            start, end = (self.end, self.start) if self.start > self.end else (self.start, self.end)

            values = [
                ValueNormalizer(start, end).denormalize(normalized_value)
                for normalized_value in Progression.get_n_normalized_values(self.n, self.rate_function)
            ]

            # If limits are switched, we need to recalculate it
            values = [
                self.end - value + self.start
                for value in values
            ] if self.start > self.end else values

            self._values = values

        return self._values

    @property
    def start(self):
        """
        The first value of the progression, that acts as a limit.
        """
        return self._start
    
    @start.setter
    def start(self, value: float):
        if not NumberValidator.is_number(value):
            raise Exception('The provided "start" value is not a number.')

        self._values = None
        self._start = value

    @property
    def end(self):
        """
        The last value of the progression, that acts as a limit.
        """
        return self._end
    
    @end.setter
    def end(self, value: float):
        if not NumberValidator.is_number(value):
            raise Exception('The provided "end" value is not a number.')
        
        self._values = None
        self._end = value

    @property
    def n(self):
        """
        The amount of total values in the progression including
        'start' and 'end'.
        """
        return self._n
    
    @n.setter
    def n(self, value: int):
        """
        The amount of total values in the progression including
        'start' and 'end'.
        """
        if not NumberValidator.is_positive_number(value) or not NumberValidator.is_int(value):
            raise Exception('The provided "n" value must be a positive and int value.')
        
        self._values = None
        self._n = value

    @property
    def rate_function(self):
        """
        The rate function that will be applied to calculate
        the 'n' values between 'start' and 'end'.
        """
        return self._rate_function
    
    @rate_function.setter
    def rate_function(self, value: RateFunctionArgument):
        # TODO: Validate rate function

        self._values = None
        self._rate_function = value

    def __init__(
        self,
        start: float,
        end: float,
        n: int,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        if not NumberValidator.is_number(start) or not NumberValidator.is_number(end):
            raise Exception('The provided "start" and "end" parameter must be numbers.')

        if not NumberValidator.is_positive_number(n) or n < 2:
            raise Exception('The provided "n" parameter is not a positive number greater or equal than 2.')
        
        self._start = start
        self._end = end
        self._n = n
        self._rate_function = rate_function
        self._values = None

    @staticmethod
    def get_n_equally_distributed_normalized_values(
        n: int
    ):
        """
        Get a list containing 'n' equally distributed (and
        normalized) values between the 'start' and 'end'
        (included). The 'n' value must be equal or greater
        than 2.

        If n = 6:
        - [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]
        """
        if not NumberValidator.is_positive_number(n) or n < 2:
            raise Exception('The provided "n" parameter is not a positive number greater or equal than 2.')

        return np.linspace(0.0, 1.0, n).tolist() if n > 0 else []

    @staticmethod
    def get_n_normalized_values(
        n: int,
        rate_function: RateFunctionArgument = RateFunctionArgument.default()
    ):
        """
        Get a list containing 0.0 and 1.0 limits and 'n' - 2
        values in between those 'start' and 'end' limits
        according to the provided 'rate_function'.

        If n = 6 and rate_function = ease_in_quad:
        - [0.0, 0.04, 0.16, 0.36, 0.64, 1.0]
        """
        if not NumberValidator.is_positive_number(n) or not n >= 2:
            raise Exception('The provided "n" parameter is not a positive number greater or equal than 2.')
        
        return [
            rate_function.get_n_value(normalized_value)
            for normalized_value in Progression.get_n_equally_distributed_normalized_values(n)
        ]

# TODO: I need to review the GraphicInterpolation and refactor
# it with the new NormalizedValue, ValueNormalizer and more...
# Code must be completely clear and as self-descriptive as
# possible


from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.text.transformer import remove_marks_and_accents


class TextFinderMode(Enum):
    """
    This is the mode in which we will look for the terms
    in the given segment text to find any coincidences.
    """

    EXACT = 'exact'
    """
    The term found must be exactly matched on the text,
    which means that accents and punctuation marks will
    be considered.
    """
    IGNORE_CASE_AND_ACCENTS = 'ignore_case_and_accents'
    """
    The term found must match, in lower case and ignoring
    the accents, the text.
    """

class TextFinder:
    """
    Class to simplify the way we look for texts
    within other texts.
    """

    def find_in_text(
        term: str,
        text: str,
        mode: TextFinderMode = TextFinderMode.IGNORE_CASE_AND_ACCENTS
    ):
        """
        Find the provided 'term' in the also provided
        'text' and obtain the start and end indexes
        of the 'term' according to that 'text'. The 
        term can be more than once in the 'text'.

        TODO: Add an ilustrative example.

        This method returns an array containing tuples
        with the start and end indexes of the term
        positions in which it has been found.
        """
        if not PythonValidator.is_string(term):
            raise Exception('The provided "term" is not a valid string.')
        
        if not PythonValidator.is_string(text):
            raise Exception('The provided "text" is not a valid string.')

        mode = TextFinderMode.to_enum(mode)

        if mode == TextFinderMode.IGNORE_CASE_AND_ACCENTS:
            term = remove_marks_and_accents(term).lower()
            text = remove_marks_and_accents(text).lower()

        text_words = text.split()
        term_words = term.split()
        
        # Store first and last index if found
        return [
            (i, i + len(term_words) - 1)
            for i in range(len(text_words) - len(term_words) + 1)
            if text_words[i:i + len(term_words)] == term_words
        ]


from setuptools import setup, find_packages


VERSION = '0.0.187'
DESCRIPTION = 'Youtube Autónomo general utils are here.'
LONG_DESCRIPTION = 'These are the general utils we need in the Youtube Autónomo project to work in a better way.'

setup(
    name = "yta_general_utils", 
    version = VERSION,
    author = "Daniel Alcalá",
    author_email = "<danielalcalavalera@gmail.com>",
    description = DESCRIPTION,
    long_description = LONG_DESCRIPTION,
    packages = find_packages(),
    install_requires = [
        'moviepy',
        'pillow',
        'backgroundremover',
        'scikit-image',
        'opencv-python',
        'numpy',
        'pyttsx3',
        'gtts',
        'pydub',
        'torch',
        'torchaudio',
        'deepfilternet'
    ],
    
    keywords = [
        'youtube autonomo general utils'
    ],
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


from yta_general_utils.web.scrapper.chrome_scrapper import google_translate
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from typing import Union


class GoogleTranslatorLanguage(Enum):
    """
    The language for the GoogleTranslator platform.
    """

    # TODO: Implement more languages
    ENGLISH = 'en'
    SPANISH = 'es'

# TODO: Create class GoogleTranslator to wrap this
def translate_text(
    text: str,
    input_language: Union[GoogleTranslatorLanguage, str] = GoogleTranslatorLanguage.ENGLISH,
    output_language: Union[GoogleTranslatorLanguage, str] = GoogleTranslatorLanguage.SPANISH
):
    """
    Returns the provided 'text' translated into the
    'output_language' using Google Traductor by chromedriver
    navigation.
    """
    if not PythonValidator.is_string(text):
        print('The provided "text" is not a valid string.')

    input_language = GoogleTranslatorLanguage.to_enum(input_language)
    output_language = GoogleTranslatorLanguage.to_enum(output_language)

    return google_translate(text, input_language.value, output_language.value)
